import pybindgen


def generate(file_):
    mod = pybindgen.Module('_rabinkarprh')
    mod.add_include('"rabinkarp.h"')
    mod.add_container('std::list<unsigned int>', 'unsigned int', 'list')
    mod.add_container('std::list<double>', 'double', 'list')

    cls = mod.add_class('RabinKarpHash')
    cls.add_constructor([pybindgen.param('int', 'my_window_size'),
                         pybindgen.param('int', 'seed')])
    cls.add_method('set_threshold',
                   None,
                   [pybindgen.param('double', 'my_threshold')])
    cls.add_method('next_chunk_boundaries',
                   pybindgen.retval('std::list<unsigned int>'),
                   [pybindgen.param('std::string*', 'str'),
                    pybindgen.param('unsigned int', 'prepend_bytes')])

    cls = mod.add_class('RabinKarpMultiThresholdHash')
    cls.add_constructor([pybindgen.param('int', 'my_window_size'),
                         pybindgen.param('int', 'seed'),
                         pybindgen.param('std::list<double>', 'my_thresholds')])
    cls.add_method('next_chunk_boundaries_with_thresholds',
                   pybindgen.retval('std::list<unsigned int>'),
                   [pybindgen.param('std::string*', 'str'),
                    pybindgen.param('unsigned int', 'prepend_bytes')])

    mod.generate(file_)



"""Fast and easy-to-use string chunking algorithms.

`fastchunking` provides two public classes meant to be used by end users.

* :class:`.SC`: Static chunking strategy.

* :class:`.RabinKarpCDC`: Rabin-Karp-based content-defined chunking strategy.

See below for details.
"""
import abc
import fastchunking._rabinkarprh as _rabinkarprh

__version__ = '0.0.3'


class BaseChunkingStrategy(object):

    """Abstract base class for chunking strategies."""
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self.window_size = 1

    @abc.abstractmethod
    def create_chunker(self, chunk_size):
        """Abstract interface for chunker creation."""
        raise NotImplementedError

    def create_multilevel_chunker(self, chunk_sizes):
        """Create a multi-level chunker performing chunking with different
        chunk sizes.

        Args:
            chunk_sizes (list): List of target chunk sizes.

                Warning:
                    For performance reasons, behavior is only defined if chunk sizes
                    are passed in order, i.e., from lowest to highest value.

        Returns:
            BaseMultiLevelChunker: A multi-level chunker object.
        """
        return DefaultMultiLevelChunker(chunk_sizes, self.create_chunker)


class BaseChunker(object):

    """Abstract class specifying the interface of chunkers."""
    __metaclass__ = abc.ABCMeta

    def next_chunk_boundaries(self, buf, prepend_bytes=0):
        """Computes the next chunk boundaries within `buf`.

        Note:
            If called more than once, output depends on `all` previous calls of this
            function: The chunking algorithm is applied to the concatenation of all
            `buf` values.

        Args:
            buf (string): The message that is to be chunked.
            prepend_bytes (Optional[int]): Optional number of zero bytes that should be
                input to the chunking algorithm before `buf`.

        Returns:
            list: List of chunk boundary positions relative to `buf`.
        """
        raise NotImplementedError


class BaseMultiLevelChunker(BaseChunker):

    """Abstract class specifying the interface of multi-level chunkers."""
    __metaclass__ = abc.ABCMeta

    def next_chunk_boundaries(self, buf, prepend_bytes=0):
        """Computes the next chunk boundaries within `buf`.

        See :meth:`.BaseChunker.next_chunk_boundaries`.
        """
        return [boundary for (boundary, _) in
                self.next_chunk_boundaries_levels(buf, prepend_bytes)]

    def next_chunk_boundaries_levels(self, buf, prepend_bytes=0):
        """Computes the next chunk boundaries within `buf`.

        Similar to :meth:`.next_chunk_boundaries`, but information about which
        chunker led to a respective boundary is included in the returned value.

        Args:
            buf (string): The message that is to be chunked.
            prepend_bytes (Optional[int]): Optional number of zero bytes that
                should be input to the chunking algorithm before `buf`.

        Returns:
            list: List of tuples (boundary, level), where boundary is a boundary
            position relative to `buf` and level is the index of the chunker
            (i.e., the index of its chunk size specified during
            instantiation) that yielded the boundary.

            If multiple chunkers yield the same boundary, it is returned
            only once, along with the highest matching chunker index.
        """
        raise NotImplementedError


class DefaultMultiLevelChunker(BaseMultiLevelChunker):

    """Default multi-level chunker implementation, turning a standard chunker
    into a multi-level chunker.

    Multi-level chunkers perform chunking using multiple chunkers of type
    :class:`.BaseChunker` with different chunk sizes in parallel.
    """

    def __init__(self, chunk_sizes, chunker_create_fn):
        # create a chunker for each chunk size
        self._chunkers = [
            chunker_create_fn(chunk_size) for chunk_size in chunk_sizes]

    def next_chunk_boundaries_levels(self, buf, prepend_bytes=0):
        """Computes the next chunk boundaries within `buf`.

        Similar to :meth:`.next_chunk_boundaries`, but information about which
        chunker led to a respective boundary is included in the returned value.

        Args:
            buf (string): The message that is to be chunked.
            prepend_bytes (Optional[int]): Optional number of zero bytes that
                should be input to the chunking algorithm before `buf`.

        Returns:
            list: List of tuples (boundary, level), where boundary is a boundary
            position relative to `buf` and level is the index of the chunker
            (i.e., the index of its chunk size specified during
            instantiation) that yielded the boundary.

            If multiple chunkers yield the same boundary, it is returned
            only once, along with the highest matching chunker index.
        """
        boundaries = {}
        for level_index, chunker in enumerate(self._chunkers):
            boundaries.update(dict(
                [(boundary, level_index) for boundary in
                 chunker.next_chunk_boundaries(buf, prepend_bytes)]))
        return sorted(boundaries.items())


class SC(BaseChunkingStrategy):

    """Static chunking strategy.

    Generates fixed-size chunks.
    """

    def create_chunker(self, chunk_size):
        """Create a chunker performing static chunking (SC) with a specific
        chunk size.

        Args:
            chunk_size (int): Target chunk size.

        Returns:
            BaseChunker: A chunker object.
        """
        return SC._Chunker(chunk_size)

    class _Chunker(BaseChunker):

        """Static chunker instance."""

        def __init__(self, chunk_size):
            self._chunk_size = chunk_size
            self._next_chunk_boundary = self._chunk_size

        def next_chunk_boundaries(self, buf, prepend_bytes=0):
            # consider prepend_bytes
            self._next_chunk_boundary = (
                (self._next_chunk_boundary - prepend_bytes) % self._chunk_size)
            if self._next_chunk_boundary == 0:
                self._next_chunk_boundary = self._chunk_size

            # determine chunk boundaries
            buf_length = len(buf)
            chunk_boundaries = range(
                self._next_chunk_boundary, buf_length + 1, self._chunk_size)

            # update next chunk boundary position
            self._next_chunk_boundary = (
                self._next_chunk_boundary - buf_length) % self._chunk_size
            if self._next_chunk_boundary == 0:
                self._next_chunk_boundary = self._chunk_size

            return list(chunk_boundaries)


class RabinKarpCDC(BaseChunkingStrategy):

    """Content-defined chunking strategy based on Rabin Karp.

    Generates variable-size chunks.
    """

    def __init__(self, window_size, seed):
        super(RabinKarpCDC, self).__init__()
        self.window_size = window_size
        self._seed = seed

    def create_chunker(self, chunk_size):
        """Create a chunker performing content-defined chunking (CDC) using
        Rabin Karp's rolling hash scheme with a specific, expected chunk size.

        Args:
            chunk_size (int): (Expected) target chunk size.

        Returns:
            BaseChunker: A chunker object.
        """
        rolling_hash = _rabinkarprh.RabinKarpHash(self.window_size, self._seed)
        rolling_hash.set_threshold(1.0 / chunk_size)
        return RabinKarpCDC._Chunker(rolling_hash)

    def create_multilevel_chunker(self, chunk_sizes):
        """Create a multi-level chunker performing content-defined chunking
        (CDC) using Rabin Karp's rolling hash scheme with different specific,
        expected chunk sizes.

        Args:
            chunk_sizes (list): List of (expected) target chunk sizes.

                Warning:
                    For performance reasons, behavior is only defined if chunk sizes
                    are passed in order, i.e., from lowest to highest value.

        Returns:
            BaseMultiLevelChunker: A multi-level chunker object.
        """
        rolling_hash = _rabinkarprh.RabinKarpMultiThresholdHash(
            self.window_size, self._seed, [1.0 / chunk_size for chunk_size in chunk_sizes])
        return RabinKarpCDC._MultiLevelChunker(rolling_hash)

    class _Chunker(BaseChunker):

        def __init__(self, rolling_hash):
            self._rolling_hash = rolling_hash

        def next_chunk_boundaries(self, buf, prepend_bytes=0):
            return list(self._rolling_hash.next_chunk_boundaries(buf, prepend_bytes))

    class _MultiLevelChunker(BaseMultiLevelChunker):

        def __init__(self, rolling_hash):
            self._rolling_hash = rolling_hash

        def next_chunk_boundaries_levels(self, buf, prepend_bytes=0):
            i = iter(list(
                self._rolling_hash.next_chunk_boundaries_with_thresholds(buf, prepend_bytes)))
            return zip(i, i)



import os
import sys
import unittest

sys.path.insert(0, os.path.abspath('..'))
import fastchunking


class StaticChunkingTests(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(StaticChunkingTests, self).__init__(*args, **kwargs)
        self.chunking_strategy = fastchunking.SC()

    def test_chunk_size_1(self):
        chunker = self.chunking_strategy.create_chunker(chunk_size=1)
        self.assertEqual(
            list(chunker.next_chunk_boundaries('0' * 3)), [1, 2, 3])
        self.assertEqual(list(chunker.next_chunk_boundaries('0' * 1)), [1])

    def test_chunk_size_2(self):
        chunker = self.chunking_strategy.create_chunker(chunk_size=2)
        self.assertEqual(list(chunker.next_chunk_boundaries('0' * 4)), [2, 4])
        self.assertEqual(list(chunker.next_chunk_boundaries('0' * 1)), [])
        self.assertEqual(list(chunker.next_chunk_boundaries('0' * 1)), [1])
        self.assertEqual(list(chunker.next_chunk_boundaries('0' * 3)), [2])
        self.assertEqual(list(chunker.next_chunk_boundaries('0' * 3)), [1, 3])

    def test_chunk_size_3(self):
        chunker = self.chunking_strategy.create_chunker(chunk_size=3)
        self.assertEqual(
            list(chunker.next_chunk_boundaries('0' * 9)), [3, 6, 9])
        self.assertEqual(
            list(chunker.next_chunk_boundaries('0' * 10)), [3, 6, 9])
        self.assertEqual(list(chunker.next_chunk_boundaries('0' * 2)), [2])
        self.assertEqual(
            list(chunker.next_chunk_boundaries('0' * 11)), [3, 6, 9])
        self.assertEqual(list(chunker.next_chunk_boundaries('0' * 1)), [1])
        self.assertEqual(list(chunker.next_chunk_boundaries('0' * 0)), [])
        self.assertEqual(list(chunker.next_chunk_boundaries('0' * 0)), [])
        self.assertEqual(list(chunker.next_chunk_boundaries('0' * 0)), [])
        self.assertEqual(list(chunker.next_chunk_boundaries('0' * 3)), [3])

    def test_chunk_size_4(self):
        chunker = self.chunking_strategy.create_chunker(chunk_size=4)
        self.assertEqual(
            list(chunker.next_chunk_boundaries('0' * 12)), [4, 8, 12])
        self.assertEqual(list(chunker.next_chunk_boundaries('0' * 2)), [])
        self.assertEqual(
            list(chunker.next_chunk_boundaries('0' * 12)), [2, 6, 10])
        self.assertEqual(list(chunker.next_chunk_boundaries('0' * 1)), [])
        self.assertEqual(
            list(chunker.next_chunk_boundaries('0' * 12)), [1, 5, 9])
        self.assertEqual(list(chunker.next_chunk_boundaries('0' * 1)), [1])

    def test_multilevel(self):
        chunker = self.chunking_strategy.create_multilevel_chunker(
            [5, 10])
        self.assertEqual(list(chunker.next_chunk_boundaries_levels(
            '0' * 20)), [(5, 0), (10, 1), (15, 0), (20, 1)])
        self.assertEqual(list(chunker.next_chunk_boundaries_levels(
            '0' * 21)), [(5, 0), (10, 1), (15, 0), (20, 1)])
        self.assertEqual(list(chunker.next_chunk_boundaries_levels(
            '0' * 22)), [(4, 0), (9, 1), (14, 0), (19, 1)])
        self.assertEqual(list(chunker.next_chunk_boundaries_levels(
            '0' * 22)), [(2, 0), (7, 1), (12, 0), (17, 1), (22, 0)])

    def test_multilevel_without_levels(self):
        chunker = self.chunking_strategy.create_multilevel_chunker(
            [5, 10])
        self.assertEqual(list(chunker.next_chunk_boundaries(
            '0' * 20)), [5, 10, 15, 20])
        self.assertEqual(list(chunker.next_chunk_boundaries(
            '0' * 21)), [5, 10, 15, 20])
        self.assertEqual(list(chunker.next_chunk_boundaries(
            '0' * 22)), [4, 9, 14, 19])
        self.assertEqual(list(chunker.next_chunk_boundaries(
            '0' * 22)), [2, 7, 12, 17, 22])

    def test_prepending(self):
        for _ in range(1024):
            content = os.urandom(1024)

            chunker = self.chunking_strategy.create_chunker(
                chunk_size=64)
            boundaries = chunker.next_chunk_boundaries(b'\0' + content)

            prepend_chunker = self.chunking_strategy.create_chunker(
                chunk_size=64)
            prepend_boundaries = prepend_chunker.next_chunk_boundaries(
                content, 1)

            self.assertEqual(
                boundaries, list(map(lambda x: x + 1, prepend_boundaries)))


class RabinKarpTests(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(RabinKarpTests, self).__init__(*args, **kwargs)
        self.chunking_strategy = fastchunking.RabinKarpCDC(48, 0)

    def test_deterministic_chunking(self):
        content = os.urandom(1024 * 1024)

        chunker = self.chunking_strategy.create_chunker(chunk_size=128)
        boundaries = chunker.next_chunk_boundaries(content)

        chunker2 = self.chunking_strategy.create_chunker(chunk_size=128)
        boundaries2 = chunker2.next_chunk_boundaries(content)

        self.assertEqual(boundaries, boundaries2)

    def test_consistent_chunking(self):
        chunker = self.chunking_strategy.create_chunker(chunk_size=128)

        part_len = 10 * 1024
        content = os.urandom(part_len)
        boundaries = chunker.next_chunk_boundaries(content + content)

        for boundary in boundaries:
            if boundary < part_len:
                self.assertIn(boundary + part_len, boundaries)

    def test_prepending(self):
        for _ in range(1024):
            content = os.urandom(1024)

            chunker = self.chunking_strategy.create_chunker(
                chunk_size=64)
            boundaries = chunker.next_chunk_boundaries(b'\0' + content)

            prepend_chunker = self.chunking_strategy.create_chunker(
                chunk_size=64)
            prepend_boundaries = prepend_chunker.next_chunk_boundaries(
                content, 1)

            self.assertEqual(
                boundaries, list(map(lambda x: x + 1, prepend_boundaries)))

    def test_sample_data_1(self):
        content = 'Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet. Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet.'

        chunker = self.chunking_strategy.create_chunker(chunk_size=128)
        boundaries = chunker.next_chunk_boundaries(content)

        self.assertEqual(boundaries, [91, 121, 387, 417])

    def test_sample_data_2(self):
        content = 'Lorem ipsum dolor sit amet, dictas definitiones ea nam, per at fugit voluptaria. Brute luptatum recusabo ne per, mei modo consul indoctum ex. Quem accusamus an sea. Graece oportere dignissim eos et, an diam voluptatibus est. Dictas aperiam est at, nibh tritani has ex, his decore aliquid ut. Ius ei iusto ludus clita, ea per inermis probatus forensibus. Eum ex ludus nullam persequeris, mel gubergren reprehendunt ad, ne mel regione disputationi. Aliquando forensibus sit ne, sea et graece causae fabulas. Vel eu pericula intellegat rationibus, in qui exerci adversarium. Ea nec feugiat placerat, eos dicat invidunt maluisset ea. Graece convenire.'

        chunker = self.chunking_strategy.create_chunker(chunk_size=16)
        boundaries = chunker.next_chunk_boundaries(content)

        self.assertEqual(boundaries, [56, 98, 119, 182, 198, 204, 214, 245, 270, 282, 287, 312, 313, 315, 317, 328, 331,
                                      345, 367, 377, 397, 410, 417, 418, 437, 443, 459, 466, 474, 475, 492, 497, 501, 522, 532, 545, 577, 597, 598, 606])

    def test_multilevel(self):
        content = 'Lorem ipsum dolor sit amet, dictas definitiones ea nam, per at fugit voluptaria. Brute luptatum recusabo ne per, mei modo consul indoctum ex. Quem accusamus an sea. Graece oportere dignissim eos et, an diam voluptatibus est. Dictas aperiam est at, nibh tritani has ex, his decore aliquid ut. Ius ei iusto ludus clita, ea per inermis probatus forensibus. Eum ex ludus nullam persequeris, mel gubergren reprehendunt ad, ne mel regione disputationi. Aliquando forensibus sit ne, sea et graece causae fabulas. Vel eu pericula intellegat rationibus, in qui exerci adversarium. Ea nec feugiat placerat, eos dicat invidunt maluisset ea. Graece convenire.'

        chunker = self.chunking_strategy.create_multilevel_chunker(
            [16, 32, 64])
        boundaries_with_levels = chunker.next_chunk_boundaries_levels(
            content)

        self.assertEqual(list(boundaries_with_levels), [(56, 0), (98, 1), (106, 0), (136, 0), (182, 1), (196, 0), (198, 2), (204, 1), (206, 0), (213, 0), (227, 0), (245, 2), (270, 1), (282, 0), (287, 0), (312, 1), (313, 1), (315, 1), (317, 0), (328, 0), (
            331, 0), (345, 0), (367, 0), (377, 2), (383, 1), (391, 0), (408, 0), (410, 2), (437, 0), (443, 0), (459, 1), (463, 0), (466, 2), (474, 1), (492, 2), (497, 2), (501, 1), (522, 2), (532, 1), (545, 1), (577, 0), (597, 0), (598, 2), (606, 0)])


class AbstractTests(unittest.TestCase):

    def test_chunking_strategy(self):
        class Test(fastchunking.BaseChunkingStrategy):

            def create_chunker(self, *args, **kwargs):
                return super(Test, self).create_chunker(*args, **kwargs)

        self.assertRaises(NotImplementedError, Test().create_chunker, 4096)
        self.assertRaises(
            NotImplementedError, Test().create_multilevel_chunker, [4096])

    def test_chunker(self):
        class Test(fastchunking.BaseChunker):

            def next_chunk_boundaries(self, *args, **kwargs):
                return super(Test, self).next_chunk_boundaries(*args, **kwargs)

        self.assertRaises(
            NotImplementedError, Test().next_chunk_boundaries, '0')

if __name__ == "__main__":
    unittest.main()



import timeit
import os
import time

import fastchunking

if __name__ == '__main__':
    print("Benchmarking RabinKarpChunking creation time...")
    NUMBER = 10000
    total_time = timeit.timeit(
        "fastchunking.RabinKarpCDC(48, 0).create_chunker(128)",
        setup="import fastchunking",
        number=NUMBER)
    print("average creation time: {:f}s\n".format(total_time / NUMBER))

    print("Benchmarking RabinKarpChunking chunking throughput...")
    SIZE = 100 * 1024 * 1024  # 100 MiB
    for chunksize in [64, 128, 256, 512, 1024, 2048, 4096, 8192, 16384, 32768]:
        chunker = fastchunking.RabinKarpCDC(
            48, 0).create_chunker(chunksize)
        content = os.urandom(SIZE)
        t = time.time()
        chunker.next_chunk_boundaries(content, 0)
        print("chunking throughput (chunksize = {} bytes): {} MiB/s".format(
            chunksize, SIZE / 1024 / 1024 / (time.time() - t)))



import codecs
import os
import re
import sys

from setuptools import setup, Extension
from setuptools.command.test import test as TestCommand

# Some general-purpose code stolen from
# https://github.com/jeffknupp/sandman/blob/5c4b7074e8ba5a60b00659760e222c57ad24ef91/setup.py

here = os.path.abspath(os.path.dirname(__file__))


class Tox(TestCommand):

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import tox
        errcode = tox.cmdline(self.test_args)
        sys.exit(errcode)


def read(*parts):
    # intentionally *not* adding an encoding option to open
    return codecs.open(os.path.join(here, *parts), 'r').read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

# Make sure build path exists.
build_path = os.path.join(here, 'build')
if not os.path.exists(build_path):
    os.mkdir(build_path)

# Generate Python bindings for bundled C++ library.
module_fname = os.path.join(build_path, "rabinkarprh.cpp")

try:
    import pybindgen #@UnusedImport
except ImportError:
    print("WARNING: Failed to import pybindgen. If you called setup.py egg_info,"
          "this is probably acceptable; otherwise, build will fail."
          "You can resolve this problem by installing pybindgen beforehand.")
else:
    with open(module_fname, "wt") as file_:
        print("Generating file {}".format(module_fname))
        from lib.rabinkarp_gen import generate
        generate(file_)

setup(
    name='fastchunking',
    version=find_version('fastchunking', '__init__.py'),

    description='Fast chunking library.',
    long_description=read('README.rst'),

    url='https://github.com/netleibi/fastchunking',

    author='Dominik Leibenger',
    author_email='python-fastchunking@mails.dominik-leibenger.de',

    license='Apache Software License',

    classifiers=[
            'Development Status :: 2 - Pre-Alpha',

            'Intended Audience :: Developers',
            'Topic :: Software Development :: Libraries :: Python Modules',

            'License :: OSI Approved :: Apache Software License',

            'Operating System :: OS Independent',

            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5'
    ],

    keywords=['text chunking', 'SC', 'static chunking', 'CDC', 'content-defined chunking', 'ML-*', 'multi-level chunking', 'ML-SC', 'ML-CDC', 'Rabin Karp', 'rolling hash'],

    packages=['fastchunking', 'lib'],

    setup_requires=['pybindgen'],
    install_requires=['pybindgen'],

    ext_modules=[
        Extension('fastchunking._rabinkarprh',
                  sources=[module_fname, 'lib/rabinkarp.cpp'],
                  include_dirs=['lib']
                  )
    ],

    test_suite='fastchunking.test',
    tests_require=['tox'],
    cmdclass={'test': Tox}
)
